// loveServer.cc

// Author: Jim Kowalkowski
// Date: 7/94
// Changes for MPF: Marty Kraimer
// 27AUG98


extern "C" {
#include <vxWorks.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
#include <taskLib.h>
#include <semLib.h>
}
 
#include "Message.h"
#include "Int32Message.h"
#include "Char8ArrayMessage.h"
#include "loveServer.h"
#include "WatchDog.h"
#include "SerialPort.h"
 
static const int readBufSize = 20;

int loveServerDebug=1;
 
enum readState {readIdle, readStart, readData};

class Love
{
public:
    Love(const char *name,const char *portName);
    static void loveServer(Love *);
    bool getStartOk() const {return(startOk);}
private:
    void talk(Int32Message *pmessage);
    static int setUpRead(char* msg,unsigned char addr,short cmd);
    static int checkSumFailure(unsigned char* msg,int response_len);
    static unsigned long getValue(char* data,int pos);
    const char *serverName;
    static byteHandlerRC byteHandler(void*,unsigned char);
    MessageServer *pMessageServer;
    SerialPort* pSerialPort;
    SEM_ID done;
    bool startOk;
    readState state;
    Int32Message  *psend;
    int nextRead;
    unsigned char readBuf[readBufSize];
};

enum { STX=0x02, ETX=0x03, ACK=0x06 };

static char codes[] = "0123456789ABCDEF";
static int hex[] = { 0,1,2,3,4,5,6,7,8,9,0,0,0,0,0,0,0,10,11,12,13,14,15,16 };

/*
    message formats:
                                   user stuff
                | header        | checksum data (ascii) | tail
    -----------------------------------------------------------
    error	| <stx> <Baddr> | N <error code>        | <ack>
    instr	| <stx> <Baddr> | <addr> <data>         | <ack>
    host	| <stx> <Baddr> | <addr> <data>         | <cksm> <etx>

    <stx>=02 <ack>=06 <etx>=03
    <cksm>=lower 8 bits of sum of checksum data
    <Baddr>=binary address of destination 0x01 to 0xff, 0x00 invalid

    all user data encoded as ascii hex - 0x12fd sent as 31 32 46 44 (4 bytes)
    ?????? should error be
    error	| <stx> <Baddr> | <addr> N <error code>        | <ack>
*/

// error codes:
#define ERR_UNDEFINED_CMD1	01
#define ERR_CHECK_SUM		02
#define ERR_CMD_NOT_PERFORMED	03
#define ERR_BAD_ASCII_CHAR_RECV 04
#define ERR_DATA_FIELD_ERROR	05
#define ERR_UNDEFINED_CMD2	06
#define ERR_HARDWARE_FAULT1	08
#define ERR_HARDWARE_FAULT2	09
#define ERR_UNDEFINED_CMD3	10

static char taskname[] = "love";
int initLoveServer(char *serverName,char *portName)
{
    Love *pLove = new Love(serverName,portName);
    if(!pLove->getStartOk()) return(0);
    int taskId = taskSpawn(taskname,100,VX_FP_TASK,2000,
	(FUNCPTR)Love::loveServer,(int)pLove,0,0,0,0,0,0,0,0,0);
    if(taskId==ERROR) printf("loveServer taskSpawn Failure\n");
    return(0);
}

Love:: Love(const char *name,const char *portName)
:   serverName(name), startOk(false), state(readIdle), psend(0), nextRead(0)
  
{
    pSerialPort = SerialPort::bind(portName,this,byteHandler);
    if(pSerialPort==0) {
	printf("%s loveServer: could not bind to SerialPort\n",serverName);
        return;
    }
    done = semBCreate(SEM_Q_FIFO,SEM_EMPTY);
    if(done==NULL) {
	 printf("%s loveServer: semBCreate failed\n",serverName);
	 return;
    }
    pMessageServer = new MessageServer(name);
    startOk = true;
}

byteHandlerRC Love::byteHandler(void* v, unsigned char data)
{
    Love *pLove = (Love *)v;
    switch(pLove->state) {
    case readIdle:
	return(byteHandlerError);
    case readStart:
	if(data!=STX) break; // just skip character
	pLove->state = readData;
	//break left out on purpose so that character is accepted
    case readData:
	if(data==ACK) {
	    pLove->state=readIdle;
	    return(byteHandlerEndRead);
	}
        if((pLove->nextRead)>=readBufSize) {
	    pLove->state=readIdle;
	    return(byteHandlerError);
        }
	pLove->readBuf[pLove->nextRead++] = data;
	break;
    default:
	logMsg("love byte_handler illegal state\n");
	pLove->state=readIdle;
    }
    return(byteHandlerOK);
}

void Love::loveServer(Love *pLove)
{
    while(true) {
        pLove->pMessageServer->waitForMessage();
	Message *message;
        while((message = pLove->pMessageServer->receive())) {
            if(message->getType()!=messageTypeInt32) {
                printf("%s loveServer got illegal message type %d\n",
		    pLove->serverName, message->getType());
		continue;
            }
            Int32Message  *preceive = (Int32Message *)message;
	    pLove->psend = (Int32Message *)pLove->pMessageServer->
		allocReplyMessage(preceive,messageTypeInt32);
	    pLove->talk(preceive);
	    delete preceive;
	    pLove->pMessageServer->reply(pLove->psend);
	}
    }
    return;
}

void Love::talk(Int32Message* preceive)
{
    unsigned char out_data[50];
    int len = setUpRead((char*)out_data,preceive->address,preceive->cmd);
    nextRead = 0;
    state = readStart;
    for(int i=0; i<readBufSize; i++) readBuf[i] = 0;
    int status = pSerialPort->writeRead(out_data,len,preceive->timeout);
    if(status!=0) {psend->status = -1; return;}
    if(readBuf[4]=='N') { 
	int error = (hex[readBuf[5]-'0'])<<4 | hex[readBuf[6]-'0'];
	printf("love: error from controller %x\n",error);
	psend->status = -1; return;
    } else {
        switch(preceive->cmd) {
        case R_PV:
            if( checkSumFailure(readBuf,8) ) {
                // checksum failure
	        if(loveServerDebug) printf("checksum failure\n");
	        psend->status = -1; return;
            } else {
                // skip over the header of 4 bytes
                unsigned long address =
                 (unsigned long)(hex[readBuf[4]-'0'])<<12|
                 (unsigned long)(hex[readBuf[5]-'0'])<<8|
                 (unsigned long)(hex[readBuf[6]-'0'])<<4|
                 (unsigned long)(hex[readBuf[7]-'0']);
                unsigned long value =
                 (unsigned long)(readBuf[8]-'0')*1000+
                 (unsigned long)(readBuf[9]-'0')*100+
                 (unsigned long)(readBuf[10]-'0')*10+
                 (unsigned long)(readBuf[11]-'0');
		int val = (long)value;
                if(address&0x01) {
                    // adjust sign
                    val *= -1;
                }
		psend->value = (int32)val;
		psend->status = 0;
            }
            break;
        case R_DPT:
            if( checkSumFailure(readBuf,2) ) {
                // checksum failure
	        if(loveServerDebug) printf("checksum failure\n");
	        psend->status = -1; return;
            } else {
                psend->value = (int32)(unsigned long)(readBuf[5]-'0');
		psend->status = 0;
            }
            break;
        default: break;
        }
    }
}

int Love::setUpRead(char* data,unsigned char addr,short cmd)
{
    int total=0;
    int pos=0;
    unsigned char sum;

    data[pos++]=STX;
    data[pos++]='L';	// binary addr, no 'L'

    // ascii version of address
    data[pos]=codes[(addr>>4)];		total+=(int)data[pos++];
    data[pos]=codes[(addr&0x0f)];	total+=(int)data[pos++];

    // command to execute
    if(cmd <= 0xff) {
        data[pos]=codes[(cmd>>4)];	total+=(int)data[pos++];
        data[pos]=codes[(cmd&0x0f)];	total+=(int)data[pos++];
    } else {
        data[pos]=codes[(cmd>>12)];		total+=(int)data[pos++];
        data[pos]=codes[(cmd>>8&0x000f)];	total+=(int)data[pos++];
        data[pos]=codes[(cmd>>4&0x000f)];	total+=(int)data[pos++];
        data[pos]=codes[(cmd&0x000f)];		total+=(int)data[pos++];
    }
    sum=(unsigned char)total; // checksum for ascii address + command
    data[pos++]=codes[(sum>>4)];
    data[pos++]=codes[(sum&0x0f)];
    data[pos++]=ETX;
    return pos;
}

int Love::checkSumFailure(unsigned char* data,int data_len)
{
    int total=0;
    int i,sum;
    unsigned char* pos=&data[1];
    // add 3 to length for addresses
    for(i=0;i<(data_len+3);i++) total+=(int)pos[i];
    total&=0x000000ff;
    sum=(hex[pos[i]-'0'])<<4 | hex[pos[i+1]-'0'];
    return (sum==total)?0:-1;
}

unsigned long Love::getValue(char* data,int pos)
{
    int i;
    unsigned long x;

    i=(pos+4); // pos is position after header
    x=(unsigned long)(data[i+0]-'0')*1000+
      (unsigned long)(data[i+1]-'0')*100+
      (unsigned long)(data[i+2]-'0')*10+
      (unsigned long)(data[i+3]-'0');

    return x;
}
