// Message.h

/* Author:  Marty Kraimer Date:    26MAR1998 */
 
/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

#ifndef MessageH
#define MessageH

#include "mpfType.h"
#include "messageType.h"
class DLList;
class DLNode;

class Message;
class ConnectMessage;
class MessagePvt;
class MessageClient;
class MessageServer;
class MessageRouterClient;
class MessageRouterServer;

typedef Message *(*createMessage)();

class MPF;

class Message
{
public:
    messageType	 getType();
    virtual ~Message() {};
    virtual void init() = 0;
    virtual int32  toBuffer(char **buffer) const;
    virtual int32  fromBuffer(const char **buffer);
    virtual int32  fromBufferSwitch(const char **buffer);
    virtual void print() const;
    static Message *allocate(messageType type);
protected:
    Message(messageType type) {this->type = type;};
private:
    messageType	type;
    MessagePvt	*pvt;
    friend class MPF;
    friend class MessageRouterClient;
    friend class MessageRouterServer;
    friend class MessageClient;
    friend class MessageServer;
    // disallow default copy constructor, assignment operator
    Message(const Message &);
    Message& operator=(const Message&);
};

inline messageType Message::getType() { return type; }

struct MessageClientPrivate;
typedef void (*clientCallback)(Message *,void *clientPvt);
typedef void (*timeoutCallback)(void *pvt);
class MessageClient
{
public:
    MessageClient(clientCallback,void *clientPvt);
    int	bind(char *server, int location);
    int	send(Message *message);
    int	sendTimeout(Message *message,int seconds,timeoutCallback,void *pvt);
    void cancelTimeout();
    void *getClientPvt();
    //Following for routers not for client
    clientCallback clientCall;
    void bindCallback(ConnectMessage *message);
    void	*routerPvt;
private:
    friend class MessageRouterClient;
    void *clientPvt;
    MessageClientPrivate *pvt;
    // disallow default  copy constructor, assignment operator
    MessageClient(const MessageClient &);
    MessageClient& operator=(const MessageClient&);
    //For now disallow destructor
    ~MessageClient();
};

inline void *MessageClient::getClientPvt() {return(clientPvt);}

struct MessageServerPrivate;
typedef Message *(*allocateMessage)();
class MessageServer
{
public:
    MessageServer(const char* name,int queueSize);
    void waitForMessage();
    Message *receive();
    Message *allocReplyMessage(Message *clientMessage,messageType);
    void  queueMessage(Message *);
    int reply(Message *);
    void setQueueSize(int size);
    const char *getName() const;
    void report() const;
private:
    friend class MessageRouterServer;
    MessageServerPrivate *pvt;
    // disallow default  copy constructor, assignment operator
    MessageServer(const MessageServer &);
    MessageServer& operator=(const MessageServer&);
    //For now disallow destructor
    ~MessageServer();
};

class MPF {

public:
  static void *allocate(DLList **ppFreeList,size_t size);
  static void free(DLList *pFreeList,void *pmessage,size_t size);
  static void switchBytes(char *to,const char *from,int nbytes);

  static int32 toBuffer(char **buffer, const char8 value);
  static int32 fromBuffer(const char **buffer,char8 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,char8 *pvalue);

  static int32 toBuffer(char **buffer, const int16 value);
  static int32 fromBuffer(const char **buffer,int16 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,int16 *pvalue);
  static int32 toBuffer(char **buffer, const int32 value);
  static int32 fromBuffer(const char **buffer,int32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,int32 *pvalue);

  static int32 toBuffer(char **buffer, const uint16 value);
  static int32 fromBuffer(const char **buffer,uint16 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,uint16 *pvalue);
  static int32 toBuffer(char **buffer, const uint32 value);
  static int32 fromBuffer(const char **buffer,uint32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,uint32 *pvalue);

  static int32 toBuffer(char **buffer,const float32 value);
  static int32 fromBuffer(const char **buffer,float32 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,float32 *pvalue);
  static int32 toBuffer(char **buffer,const float64 value);
  static int32 fromBuffer(const char **buffer,float64 *pvalue);
  static int32 fromBufferSwitch(const char **buffer,float64 *pvalue);

  static int32 toBuffer(char **buffer,const char *pstring);
  static int32 fromBuffer(const char **buffer,char *pstring,int32 maxlen);
  static int32 fromBufferSwitch(const char **buffer,
      char *pstring,int32 maxlen);

  static int32 toBuffer(char **buffer,const void *,int32 nbytes);
  static int32 fromBuffer(const char **buffer,void *,int32 nbytes);
};

inline void MPF::switchBytes(char *to,const char *from,int nbytes)
{
   for(int i=0; i<nbytes; i++) to[i] = from[nbytes-i-1];
}

inline int32 MPF::toBuffer(char **buffer, const char8 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,char8 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,char8 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const int16 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,int16 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,int16 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const int32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,int32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,int32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const uint16 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,uint16 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,uint16 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer, const uint32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,uint32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,uint32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer,const float32 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,float32 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,float32 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

inline int32 MPF::toBuffer(char **buffer,const float64 value)
{
    memcpy(*buffer,(char *)&value,sizeof(value));
    *buffer += sizeof(value);
    return(sizeof(value));
}
inline int32 MPF::fromBuffer(const char **buffer,float64 *pvalue)
{
    memcpy((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}
inline int32 MPF::fromBufferSwitch(const char **buffer,float64 *pvalue)
{
    MPF::switchBytes((char *)pvalue,*buffer,sizeof(*pvalue));
    *buffer += sizeof(*pvalue);
    return(sizeof(*pvalue));
}

#endif /* MessageH */
