// RMRServer.cc - Remote Message Router Server side

/* Author:  Marty Kraimer Date:    08APR1998 */

/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

extern "C" {
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <stdio.h>
 
#include <semLib.h>
#include <taskLib.h>
}

#include "FreeList.h"
#include "DLList.h"
#include "Message.h"
#include "MessagePvt.h"
#include "ConnectMessage.h"
#include "Tcp.h"

enum connectionState {stateStart,stateWaitConnect,stateConnected};

static const char *connectionStateString[] =
{ "stateStart","stateWaitConnect","stateConnected" };

enum messageClass {messageClassBind, messageClassUser};

static void connectCallback(void *pvt,tcpConnectState state);
static void receiveCallback(void *pvt,void *buffer,int32 nbytes);
static void RMRreplyTask(void *pvt);

struct ServerInfo {
    MessageServer *pMessageServer;
    void	*routerClientPvt;
};

struct ReplyNode {
    Message	*pmessage;
    ServerInfo	*pServerInfo;
};

struct RMRconnectMessage {
    char	magic[4];		// will be MPF
    int32	determineByteOrder;	// sent as 1
};

class RMRServer: public MessageRouterServer
{
public:
    RMRServer(int location,Tcp *pTcp,int bufsize,int queueSize);
    virtual ~RMRServer() {}
    virtual int reply(Message *message);
    virtual void report();
    //Remaining is for private use by this source module
    void	lock() {semTake(lockSem,WAIT_FOREVER);}
    void	unlock() {semGive(lockSem);}
    void	connect();
    void	disconnect();
    void	receiveConnect(void *buffer,int32 nbytes);
    void	receive(void *buffer,int32 nbytes);
    Tcp		*pTcp;	
    const int	replyBufferSize;
    bool	sameByteOrder;
    connectionState state;
    char	*replyBuffer;
    DLList	*pServerInfoList;
    DLList	*preplyList;
    FreeList	*pFreeReplyNode;
    SEM_ID	replyRequest;
    SEM_ID	lockSem; //for modifying any list;
    int		replyTaskid;
    int		queueSize;
    int		replyQueueFull;
};

int tcpMessageRouterServerStart(int location,int port,
    const char *address,int bufsize,int queueSize)
{
    if(bufsize<100) {
	printf("bufsize <100. Set = 1024\n");
	bufsize = 1024;
    }
    if(queueSize<=0) {
	printf("queueSize %d. Set = 100\n",queueSize);
	queueSize = 100;
    }
    Tcp *pTcp = new Tcp(true,port,address,bufsize);
    new RMRServer(location,pTcp,bufsize,queueSize);
    return(0);
}

static char taskname[] = "RMRServer";

RMRServer::RMRServer(int location,Tcp *ptcp,int bufsize,int queueSize)
  : MessageRouterServer(location), pTcp(ptcp), replyBufferSize(bufsize),
    sameByteOrder(false), state(stateStart),
    queueSize(queueSize), replyQueueFull(0)
{
    replyBuffer = new char[replyBufferSize];
    pServerInfoList = new DLList;
    preplyList = new DLList;
    pFreeReplyNode = new FreeList((int)sizeof(ReplyNode),100);
    replyRequest = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    lockSem = semBCreate(SEM_Q_FIFO, SEM_FULL);
    replyTaskid = taskSpawn(taskname,110,VX_FP_TASK,2000,
	(FUNCPTR)RMRreplyTask, (int)this,0,0,0,0,0,0,0,0,0);
    if(replyTaskid==ERROR) {
       printf("RMRServer taskSpawn Failure\n");
    }
    pTcp->connect((void *)this,connectCallback,receiveCallback);
    prouter->registerServerRouter(this,getLocation());
}

static void connectCallback(void *pvt,tcpConnectState state)
{
    RMRServer *pRMRServer = (RMRServer *)pvt;
    switch(state) {
        case tcpConnectYes: pRMRServer->connect(); break;
        case tcpConnectNo: pRMRServer->disconnect(); break;
        default:
	    printf("MRMServer:connectCallback illegal state %d\n",(int)state);
    }
}

static void receiveCallback(void *pvt,void *buffer,int32 nbytes)
{
    RMRServer *pRMRServer = (RMRServer *)pvt;
    pRMRServer->lock();
    connectionState state = pRMRServer->state;
    pRMRServer->unlock();
    switch(state) {
    case stateWaitConnect: pRMRServer->receiveConnect(buffer,nbytes); break;
    case stateConnected: pRMRServer->receive(buffer,nbytes); break;
    default:
	printf("tcpRouter: receiveCallback illegal state\n");
	break;
    }
}

int RMRServer::reply(Message *pmessage)
{
    MessagePvt *msgPvt = getMessagePvt(pmessage);
    if(!msgPvt) return(-1);
    if(state!=stateConnected) return(-1);
    if(preplyList->getCount() >= queueSize) {
	++replyQueueFull;
	return(-1);
    }
    ReplyNode *pReplyNode = (ReplyNode *)pFreeReplyNode->calloc();
    pReplyNode->pmessage = pmessage;
    pReplyNode->pServerInfo = (ServerInfo *)msgPvt->routerServerPvt;
    DLNode *node = DLList::allocNode();
    node->obj = (void *)pReplyNode;
    lock();
    preplyList->addEnd(node);
    unlock();
    semGive(replyRequest);
    return(0);
}

void RMRServer::report()
{
    printf(" %d RMRServer %s queueSize %d inQueue %d replyQueueFull %d\n",
	getLocation(),connectionStateString[state],
	queueSize,preplyList->getCount(),replyQueueFull);
    pTcp->printStats();
}

void RMRServer::connect()
{
    lock();
    state = stateWaitConnect;
    unlock();
}

void RMRServer::disconnect()
{
    lock();
    DLNode *node;
    while((node = pServerInfoList->getFirst())) {;
	ServerInfo *pServerInfo = (ServerInfo *)node->obj;
	delete pServerInfo;
	pServerInfoList->freeNode(node);
    }
    while((node = preplyList->getFirst())) {;
	ReplyNode *pReplyNode = (ReplyNode *)node->obj;
	Message *pmessage = pReplyNode->pmessage;
	pFreeReplyNode->free(pReplyNode);
	preplyList->freeNode(node);
	delete pmessage;
    }
    state = stateStart;
    unlock();
}

void RMRServer::receiveConnect(void *buffer,int32 nbytes)
{
    if(nbytes!=sizeof(RMRconnectMessage)) {
	printf("RMRServer::receiveConnect received illegal header\n");
	taskSuspend(0);
    }
    RMRconnectMessage *pmessage = (RMRconnectMessage *)buffer;
    if((::strncmp(pmessage->magic,"MPF",4) != 0)) {
        printf("RMRServer::receiveConnect received illegal header\n");
        taskSuspend(0);
    }
    if(pmessage->determineByteOrder==1) {
        sameByteOrder = true;
    } else {
        sameByteOrder = false;
    }
    // now send connect message back to client router
    RMRconnectMessage message;
    ::strcpy(message.magic,"MPF");
    message.determineByteOrder = 1;
    int32 status = pTcp->sendFlush(&message,sizeof(message));
    if(status) {
	printf("RMRServer::connect tcp status is %d\n",status);
	disconnect();
	return;
    }
    lock();
    state = stateConnected;
    unlock();
}

void RMRServer::receive(void *buffer,int32 totalBytes)
{
    const char *pnext = (char *)buffer;
    int32 nbytes = 0;
    int32 temp = 0;
    if(sameByteOrder) nbytes += MPF::fromBuffer(&pnext,&temp);
    else nbytes += MPF::fromBufferSwitch(&pnext,&temp);
    messageClass msgclass = (messageClass)temp;
    switch(msgclass) {
        case messageClassBind: {
	    //bind message : server,routerClientPvt
	    ServerInfo *pServerInfo = new ServerInfo;
	    ::memset((char *)pServerInfo,0,sizeof(ServerInfo));
	    //Use replyBuffer as place to store name of server
	    char *server = replyBuffer;
	    nbytes += MPF::fromBuffer(&pnext,server,(int32)replyBufferSize);
	    MessageServer *pMessageServer = prouter->findMessageServer(server);
	    nbytes += MPF::fromBuffer(&pnext,
		&pServerInfo->routerClientPvt,(int32)sizeof(void *));
	    //reply messageClassBind, routerClientPvt, pServerInfo
	    char *out = replyBuffer;
	    int32 temp = messageClassBind;
	    int32 messageLength = MPF::toBuffer(&out,temp);
	    messageLength += MPF::toBuffer(&out,
		(void *)&pServerInfo->routerClientPvt,(int32)sizeof(void *));
	    if(pMessageServer) {
	        messageLength += MPF::toBuffer(&out,(void *)&pServerInfo,
		    (int32)sizeof(void *));
		pServerInfo->pMessageServer = pMessageServer;
		// see if ServerInfo already exists
		ServerInfo *pinfo = 0;
		DLNode *node = pServerInfoList->getFirst();
		while(node) {
		    pinfo = (ServerInfo *)node->obj;
		    if(pinfo->pMessageServer==pMessageServer) break;
		    node = pServerInfoList->getNext(node);
		}
		if(node) {
		    delete pServerInfo;
		    pServerInfo = pinfo;
		} else {
		    node = DLList::allocNode();
		    node->obj = (void *)pServerInfo;
		    pServerInfoList->addEnd(node);
		}
	    } else {
		void *zero = 0;
		messageLength += MPF::toBuffer(&out,(void *)&zero,
		    (int32)sizeof(void *));
		delete pServerInfo;
	    }
	    pTcp->sendFlush(replyBuffer,messageLength);
	    break;
	}
	
        case messageClassUser: {
	    // client message: pServerInfo,routerClientPvt,message
	    ServerInfo *pServerInfo = 0;
	    nbytes += MPF::fromBuffer(&pnext,
		(void *)&pServerInfo,(int32)sizeof(void *));
	    void *routerClientPvt=0;
	    nbytes += MPF::fromBuffer(&pnext,
		&routerClientPvt,sizeof(routerClientPvt));
	    int32 nb;
	    int32 mtype;
	    if(sameByteOrder)  nb = MPF::fromBuffer(&pnext,&mtype);
	    else nb = MPF::fromBufferSwitch(&pnext,&mtype);
	    pnext -= nb;
	    Message *pmessage = Message::allocate((messageType)mtype);
	    MessagePvt *msgPvt = getMessagePvt(pmessage);
	    msgPvt->pMessageRouterServer = this;
	    msgPvt->pMessageServer = pServerInfo->pMessageServer;
	    msgPvt->routerClientPvt = routerClientPvt;
	    msgPvt->routerServerPvt = (void *)pServerInfo;
	    if(sameByteOrder) nbytes += pmessage->fromBuffer(&pnext);
	    else nbytes += pmessage->fromBufferSwitch(&pnext);
	    pServerInfo->pMessageServer->queueMessage(pmessage);
	    break;
	}
        default:
	    printf("RMRServer::receive illegal message class %d\n",
		msgclass);
            taskSuspend(0);
    }
    if(nbytes!=totalBytes) {
	printf("RMRServer::receive received incorrect message size"
	    " received %d used %d\n",totalBytes,nbytes);
    }
}

static void RMRreplyTask(void *pvt)
{
    RMRServer *pRMRServer = (RMRServer *)pvt;
    while(true) {
	semTake(pRMRServer->replyRequest,WAIT_FOREVER);
	// send each reply message
	//Each message is messageClassUser,routerClientPvt,message
	DLNode *node;
	while(true) {
	    pRMRServer->lock();
	    node = pRMRServer->preplyList->getFirst();
	    if(!node) {
		pRMRServer->unlock();
		break;
	    }
	    ReplyNode *pReplyNode = (ReplyNode *)node->obj;
	    pRMRServer->preplyList->freeNode(node);
	    Message *pmessage = pReplyNode->pmessage;
	    MessagePvt *msgPvt = pRMRServer->getMessagePvt(pmessage);
	    void *routerClientPvt = msgPvt->routerClientPvt;
	    pRMRServer->pFreeReplyNode->free(pReplyNode);
	    int32 messageLength = 0;
	    char *buffer = pRMRServer->replyBuffer;
	    int32 temp = messageClassUser;
	    messageLength += MPF::toBuffer(&buffer,temp);
	    messageLength += MPF::toBuffer(&buffer,&routerClientPvt,
		(int32)sizeof(void *));
	    messageLength += pmessage->toBuffer(&buffer);
	    pRMRServer->pTcp->send(pRMRServer->replyBuffer,messageLength);
	    delete pmessage;
	    pRMRServer->unlock();
	}
	pRMRServer->pTcp->flush();
    }
}
