/* FreeList.cc	*/

/* Author:  Marty Kraimer Date:    26MAR1998 */
 
/********************COPYRIGHT NOTIFICATION**********************************
This software was developed under a United States Government license
described on the COPYRIGHT_UniversityOfChicago file included as part
of this distribution.
****************************************************************************/

/*
 * Modification Log:
 * -----------------
 * .01  26MAR1998	mrk	Converted freeList from C to C++
*/


extern "C" {
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <taskLib.h>
#include <semLib.h>

}

#include "FreeList.h"

//adjustToWorstCaseAlignment taken from epics base
static size_t adjustToWorstCaseAlignment(size_t size);

class allocMem {
public:
    class allocMem	*next;
    void		*memory;
};

class FreeListPrivate {
public:
    int		size;
    int		nmalloc;
    void	*head;
    allocMem	*mallochead;
    size_t	nBlocksAvailable;
    SEM_ID	lock;
};

FreeList:: FreeList(int size, int nmalloc)
{
    pprivate = (FreeListPrivate *)
	::calloc((size_t)1,(size_t)sizeof(FreeListPrivate));
    if(!pprivate) {
	taskSuspend(0);
    }
    pprivate->size = adjustToWorstCaseAlignment(size);
    pprivate->nmalloc = nmalloc;
    pprivate->head = 0;
    pprivate->mallochead = 0;
    pprivate->nBlocksAvailable = 0u;
    pprivate->lock = semBCreate(SEM_Q_PRIORITY,SEM_FULL);
    return;
}

FreeList:: ~FreeList()
{
    allocMem	*phead;
    allocMem	*pnext;

    phead = pprivate->mallochead;
    while(phead) {
	pnext = phead->next;
	::free(phead->memory);
	::free(phead);
	phead = pnext;
    }
    ::free((void *)pprivate);
}

void *FreeList:: calloc()
{
    void	*ptemp;

    ptemp = malloc();
    if(ptemp) memset((char *)ptemp,0,pprivate->size);
    return(ptemp);
}

void *FreeList:: malloc()
{
    void	*ptemp;
    void	**ppnext;
    allocMem	*pallocmem = 0;
    int		i;

    semTake(pprivate->lock,WAIT_FOREVER);
    ptemp = pprivate->head;
    if(!ptemp) {
        printf("FreeList:: malloc nmalloc %d size %d\n",
            pprivate->nmalloc,pprivate->size);
	ptemp = (void *)::malloc(pprivate->nmalloc*pprivate->size);
	pallocmem = (allocMem *)::calloc(1,sizeof(allocMem));
	if(!ptemp || !pallocmem) {
	    semGive(pprivate->lock);
	    printf("FreeList::malloc could not get storage\n");
	    taskSuspend(0);
	}
	pallocmem->memory = ptemp;
	if(pprivate->mallochead)
	    pallocmem->next = pprivate->mallochead;
	pprivate->mallochead = pallocmem;
	for(i=0; i<pprivate->nmalloc; i++) {
	    ppnext = (void **)ptemp;
	    *ppnext = pprivate->head;
	    pprivate->head = ptemp;
	    ptemp = ((char *)ptemp) + pprivate->size;
	}
	ptemp = pprivate->head;
        pprivate->nBlocksAvailable += pprivate->nmalloc;
    }
    ppnext = (void **)pprivate->head;
    pprivate->head = *ppnext;
    pprivate->nBlocksAvailable--;
    semGive(pprivate->lock);
    return(ptemp);
}

void FreeList:: free(void *pmem)
{
    void	**ppnext;

    semTake(pprivate->lock,WAIT_FOREVER);
    ppnext = (void **)pmem;
    *ppnext = pprivate->head;
    pprivate->head = pmem;
    pprivate->nBlocksAvailable++;
    semGive(pprivate->lock);
}

size_t FreeList:: itemsAvailable()
{
    return pprivate->nBlocksAvailable;
}

static size_t adjustToWorstCaseAlignment(size_t size)
{
    int align_size, adjust;
    struct test_long_word { char c; long lw; };
    struct test_double { char c; double d; };
    struct test_ptr { char c; void *p; };
    int test_long_size = sizeof(struct test_long_word) - sizeof(long);
    int test_double_size = sizeof(struct test_double) - sizeof(double);
    int test_ptr_size = sizeof(struct test_ptr) - sizeof(void *);
    size_t adjusted_size = size;
    
    /*
     * Use Jeff's alignment tests to determine worst case of long,
     * double or pointer alignment requirements.
    */
    align_size = test_long_size > test_ptr_size ?
                  test_long_size : test_ptr_size;
    
    align_size = align_size > test_double_size ?
                  align_size : test_double_size;
  
    /*
     * Increase the size to fit worst case alignment if not already
     * properly aligned.
     */ 
    adjust = align_size - size%align_size;
    if (adjust != align_size) adjusted_size += adjust;
    
    return (adjusted_size);
}
